/* ****************************************** */
/* Emu75: HP-75  software emulator for PC     */
/*                                            */
/* userpc.c   user interface for emu75        */
/* system dependant (PC version)              */
/*                                            */
/* Version 1.33 J-F Garnier 2010-2022         */
/* ****************************************** */




#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <string.h>

#include "version.h"

static int adsp=0;          /* active display: 1: LCD, 2:screen */
/* DISPLAY */
static int curx, cury;      /* cursor position */
static int nlines=24;       /* number of lines */
static int ncols=80;        /* number of cols */
static char fready;         /* display ready */
static char fesc=0;         /* ESC sequence on going */
static char fcur=0;         /* cursor active */
static char fctyp=1;        /* cursor type */
/* LCD */
static int lcpos=1;        /* cursor position */
static char lctyp=1;       /* cursor type */
/* keyboard */
#ifdef KBUF
static char keybuf[2000];
static int nbkey=0;
#endif


void update_cursor(void);
extern void alpha(char *dest,int n);


#if __TURBOC__ < 0x200
#define _NOCURSOR 0x2000
#define _NORMALCURSOR 0x607
#define _SOLIDCURSOR 0x007
void _setcursortype(int n)
{
  _AX=0x0100;
  _CX=n;
  geninterrupt(0x10);
}
#endif

void restore_cursor(void)
{
  _setcursortype(_NORMALCURSOR);
}


void LCDwindow(void);



/* ------------------ timer -------------------------------- */

// check if a new 55ms tick happened
int tsttimer(void)
{
   int far *ptr;
   int n, ret;
   static int last=0;

   ptr=MK_FP(0x40,0x6c);
   n=*ptr;
   if (n==last) ret=0; else ret=1;
   last=n;
   return(ret);
}



/* ------------------ keyboard -------------------------------- */

// test if a key has been pressed
int tstkey(void)
{
   int far *ptr;
   int n;

   ptr=MK_FP(0x40,0x1a);
   n=*ptr++;
   return(n!=*ptr);
}



// read a key, translates to HP75 code
int readkey(int fwait)
{
  unsigned char c, k, f;


#ifdef KBUF
  if (!fwait)
    if (!kbhit()&&(nbkey==0)) return(0);
  while (kbhit()) {
    c=getch();
    if (nbkey<2000) {
      keybuf[nbkey++]=c;
    }
  }
  c=keybuf[0];
  nbkey--;
  for (i=0;i<nbkey;i++)
    keybuf[i]=keybuf[i+1];
#else
  if (!fwait)
    if (!kbhit()) return(0);
//    if (!tstkey()) return(0);
  c=getch();
#endif

  if (c==26) {
    // ctrl-Z: break to debugger
    return(-1);
  }

  f=0;
  k=0;
  if (c!=0) {
    if (c>127) c=0;  // remove spec. char
    switch (c) {
      case 9:  k=0x8e; break;   // TAB
// c-E:EDIT  c-F:FETCH  c-T:SST  c-R:RUN
      case 5: k=0x83; break;    // c-E: EDIT
      case 6: k=0x89; break;    // c-F: FETCH
      case 10:k=0xc9; break;    // c-ENTER: c-FETCH
      case 18:k=0x8d; break;    // c-R: RUN
      case 20:k=0xad; break;    // c-T: SST
      case 27: k=0x80; break;   // ESC = ATTN
      case 127: k=0xaa; break;  // ctrl-backspace = -LINE
      case 8:
      case 13: k=c; break;
      default: if (c>=32) k=c;
    }
  }
  else {
#ifdef KBUF
    c=keybuf[0];
    nbkey=0;
#else
    c=getch();
#endif
    // see p40,149
    switch(c) {
      case 15: k=0xae; break;     // shift TAB
      case 59: k=0x80; break;     /* F1   ATTN */
      case 60: k=0x81; break;     /* F2   TIME */
      case 61: k=0x82; break;     /* F3   APPT */
      case 62: k=0x83; break;     /* F4   EDIT */
      case 63: k=0x89; break;     /* F5   FETCH*/
      case 64: k=0x8b; break;     /* F6   CLR*/
      case 65: k=0xa8; break;     /* F7 s-I/R*/
      case 66: k=0x8d; break;     /* F8  RUN */
      case 67: k=0x01; break;     /* F9  ctrl-A*/
      case 68: k=0x02; break;     /* F10 ctrl-B*/
      case 133:k=0x03; break;     /* F11 ctrl-C*/
      case 134:k=0x04; break;     /* F12 ctrl-D*/
      case 84: k=0xa0; break;     /* sF1   OFF */
      case 85: k=0xa1; break;     /* sF2 */
      case 86: k=0xa2; break;     /* sF3 */
      case 87: k=0xa3; break;     /* sF4 */
      case 88: k=0xc9; break;     /* sF5  CMDS */
      case 89: k=0xcb; break;     /* sF6  clr scr*/
      case 90: k=0xc8; break;     /* sF7  ctrl-I/R*/
      case 91: k=0xad; break;     /* sF8   SST*/
      case 92: k=0x0e;  break;     /* sF9 */
      case 93: k=0x0f;  break;     /* sF10 */
      case 135:k=0x10;  break;     /* sF11 */
      case 136:k=0x11;  break;     /* sF12 */
      case 75: k=0x86; break;     /* <- */
      case 77: k=0x87; break;     /* -> */
      case 72: k=0x84; break;     /* ^  */
      case 80: k=0x85; break;     /* v  */
      case 71: k=0xa6; break;     /* << */
      case 79: k=0xa7; break;     /* >> */
      case 73: k=0xa4; break;     /* ^^ */
      case 81: k=0xa5; break;     /* vv */
      case 83: k=0x8a; break;     /* DEL:  -CHAR */
      case 82: k=0x88; break;     /* I/R   */
      case 118:
      case 147:k=0xaa; break;     /* ctrl-DEL: -LINE */
      case 149:
      default: k=0;
    }
    f=1;
  }

  /* if control keys (cursor, etc) :   */
  if (f) {
    /* clear buffer for controlled autorepeat of ctrl keys */
    while (kbhit()) getch();
  }

  return(k);
}


// test if a shift key is kept down
int iskeydown(void)
{
 return(bioskey(2)&3);
}

/* ----------------- display -------------------------- */


void fen_aide(int mode_printer)
{
  window(1,1,ncols,nlines); /* 25 */
  clrscr();
  window(1,1,ncols,nlines-7); /* 18 */
  gotoxy(1,2);
  textbackground(BLUE);
  textcolor(WHITE);
  /*       1234567890123456789012345678901234567890 */
  gotoxy(1,1);

  cprintf("  Emu75: HP-75C & HP-IL system emulator");
  if (ncols>40)
    cprintf("         J-F GARNIER 2010, 2022         ");
  gotoxy(1,2);
  cprintf(USER);

  /*       1234567890123456789012345678901234567890 */
  window(1,nlines-2,ncols,nlines); /*  window(1,23,80,25); */
  textattr(WHITE);
  clrscr();
  window(1,1,ncols,nlines);  /*   window(1,1,80,25); */
  textattr(LIGHTGRAY<<4);

  gotoxy( 2,nlines-3); cprintf(" f1  ");
  gotoxy( 8,nlines-3); cprintf(" f2  ");
  gotoxy(14,nlines-3); cprintf(" f3  ");
  gotoxy(20,nlines-3); cprintf(" f4  ");

  gotoxy(29,nlines-3); cprintf(" f5  ");
  gotoxy(35,nlines-3); cprintf(" f6  ");
  if (ncols>40) {
    gotoxy(41,nlines-3); cprintf(" f7  ");
    gotoxy(47,nlines-3); cprintf(" f8  ");

    gotoxy(56,nlines-3); cprintf(" f9  ");
    gotoxy(62,nlines-3); cprintf(" f10 ");
    gotoxy(68,nlines-3); cprintf(" f11 ");
    gotoxy(74,nlines-3); cprintf(" f12 ");
  }
  gotoxy( 2,nlines-2); cprintf(" Off ");
  gotoxy( 8,nlines-2); cprintf("     ");
  gotoxy(14,nlines-2); cprintf("     ");
  gotoxy(20,nlines-2); cprintf("     ");

  gotoxy(29,nlines-2); cprintf(" Last");
  gotoxy(35,nlines-2); cprintf(" Cls ");
  if (ncols>40) {
    gotoxy(41,nlines-2); cprintf(" LitU");
    gotoxy(47,nlines-2); cprintf(" SST ");

    gotoxy(56,nlines-2); cprintf("     ");
    gotoxy(62,nlines-2); cprintf("     ");
    gotoxy(68,nlines-2); cprintf("     ");
    gotoxy(74,nlines-2); cprintf("     ");
  }
  gotoxy( 2,nlines-1); cprintf(" ATTN");
  gotoxy( 8,nlines-1); cprintf(" TIME");
  gotoxy(14,nlines-1); cprintf(" APPT");
  gotoxy(20,nlines-1); cprintf(" EDIT");

  gotoxy(29,nlines-1); cprintf("FETCH");
  gotoxy(35,nlines-1); cprintf(" CLR ");
  if (ncols>40) {
    gotoxy(41,nlines-1); cprintf(" Lit ");
    gotoxy(47,nlines-1); cprintf(" RUN ");

    gotoxy(56,nlines-1); cprintf("     ");
    gotoxy(62,nlines-1); cprintf("     ");
    gotoxy(68,nlines-1); cprintf("     ");
    gotoxy(74,nlines-1); cprintf("     ");
  }
  gotoxy(1,nlines);
  textbackground(BLUE);
  textcolor(WHITE);
/*  cprintf("  ESC:ATTN  c-DEL:-LINE  c-ENTER:CMDS  c-E:EDIT  c-F:FETCH  c-T:SST  c-R:RUN "); */
  if (ncols>40)
    cprintf("   Esc:ATTN  ctrl-Del:-LINE  ctrl-Enter:Last  Home:|<  End:>|  Pg^:^  PgV:V    ");
  else
    cprintf("Esc:ATTN ctrl-Del:-LINE ctrl-Enter:Last");

  textattr(WHITE);

  gotoxy(1,nlines-4);
/*  cprintf("  <-  g  f  AC  BAT  USER  RAD  0 1 2 3 4  (o)  PRGM  SUSP  CALC  ->        * ");
  cprintf("                     AC  BAT  USER  RAD  0 1 2 3 4  (o)  PRGM  SUSP  CALC    * ");  */
  gotoxy(1,nlines-5);

  if (mode_printer) cprintf(SIG);

  curx=1;
  cury=2;
}




void prepa_ecran(int mode_printer)
{
  int i;

  gotoxy(39,1);
  cprintf("   ");
  ncols=wherex();  ncols=80;
  gotoxy(1,24);
  cprintf("\n\n\n\n");
  nlines=wherey();
  if (nlines>27) {
    for (i=0;i<10;i++) cprintf("\n\n\n\n");  // 0.97 updated for 66-line mode
    nlines=wherey();
  }

  fen_aide(mode_printer);
  fready=1;
  fesc=0;
  adsp=0;
}



void fin_ecran(void)
{
  window(1,1,ncols,nlines);
  textattr(WHITE);
  gotoxy(1,nlines);
  _setcursortype(_NORMALCURSOR);
  cprintf(" \r\n");
  fready=0;
  adsp=0;
}


void LCDwindow(void)
{
  window(1,nlines-5,ncols,nlines-4);
  textattr(WHITE);
}

void SCRwindow(void)
{
  window(1,2,ncols,nlines-7);
  textattr(BLACK*16+LIGHTGREEN);
}




void update_cursor(void)
{
  if ((lcpos!=0)&&(lctyp!=0)) {
    if (adsp!=1) {
      LCDwindow();
      adsp=1;
    }
    gotoxy(lcpos,1);
    if (lctyp==1)
      _setcursortype(_NORMALCURSOR);
    else
      _setcursortype(_SOLIDCURSOR);
  }
  else if (fcur==0) {
    _setcursortype(_NOCURSOR);
  }
  else {
    if (adsp!=2) {
      SCRwindow();
      adsp=2;
    }
    gotoxy(curx, cury);
    if (fctyp==2)
      _setcursortype(_NORMALCURSOR);
    else
    _setcursortype(_SOLIDCURSOR);
  }


}



/* lcd */

void strflags(char *bin, char *ligne)
{
/* "....,....,...x..x.,xx..xxx..xxxx..xxx...x.x.x.x.x..xxx..xxxx..xxxx..xxxx....x. */
/* ".xxxxx.xxxxx.xxxxx.xxxxx....xxxxx.xxxxx.xxxxx.xxxxx....xxxxx.xxxxx.xxxxx.xxxxx */
  strcpy(ligne,"                                                           ");
/*  strcpy(ligne,"              g  f  AC  BAT  USER  RAD  0 1 2 3 4  (o)  PRGM  SUSP  CALC    * "); */
/*  strcpy(ligne,"                    AC  BAT  USER  RAD  0 1 2 3 4  (o)  PRGM  SUSP  CALC    * "); */
/*  strcpy(ligne,"                    ERROR          PRGM     2 3 4  (o)  PRGM  SUSP  CALC    * "); */

  if (bin[0]&8)  { ligne[ 4]='B'; ligne[ 5]='A'; ligne[ 6]='T'; }  /* BAT */

  if (bin[0]&4)  { ligne[ 9]='E'; ligne[10]='R';
                   ligne[11]='R'; ligne[12]='O'; ligne[13]='R'; }  /* ERR */

  if (bin[0]&2)  { ligne[17]='P'; ligne[18]='R';
                   ligne[19]='G'; ligne[20]='M'; }                 /* PRGM */

  if (bin[0]&1)  { ligne[24]='A'; ligne[25]='P';
                   ligne[26]='P'; ligne[27]='T'; }                 /* APPT */

  if (bin[4])    { ligne[57]='*'; }                                /* * */
}



void update_display(char *buf1, char *buf2, int cpos, int ctype)
{
  char ligne[80];

  if (adsp!=1) {
    LCDwindow();
    adsp=1;
  }
  gotoxy(1,1);

  if (buf1!=NULL) {
    cputs("                                       ");
    if (ncols>40)
      cputs("                                        ");
    gotoxy(1,1);
    buf1[78]=0;  //
    if (ncols<=40) buf1[38]=0;
/* manage the ~ in TIME display */
    alpha(buf1,78);
    cputs(buf1);
  }

  if (buf2!=NULL) {
    if (ncols>40) {
      gotoxy(20,2); /* 22 */
      strflags(buf2,ligne);
      cputs(ligne);
    }
    else {
      gotoxy(7,2);
      strflags(buf2,ligne);
      cputs(ligne);
    }
  }

  if (cpos!=0) {
    lcpos=cpos;
    if (ctype&0x40)
      lctyp=0;
    else if (ctype&0x80)
      lctyp=1;
    else
      lctyp=2;
  }
  update_cursor();

}







/* display */



void clrdsp(void)
{

  if (adsp!=2) {
    SCRwindow();
    adsp=2;
  }
  clrscr();
  curx=1;
  cury=1;
  fesc=0;
  fctyp=1;
  fcur=0;

  update_cursor();
}

void disp_string(char *s)
{
  if (!fready)
    cprintf(s);
  else {
    if (adsp!=2) {
      SCRwindow();
      gotoxy(curx,cury);
      adsp=2;
    }
    cprintf("%s",s);
    curx=wherex();
    cury=wherey();
    update_cursor();
  }
}



void disp_char(int n)
{
  int i;
  static int f=0;   // flag ESC-H, ESC-J sequence (for the HP75)

  if (adsp!=2) {
    SCRwindow();
    gotoxy(curx,cury);
    adsp=2;
  }

  if (fesc) {
    switch (fesc) {
      case 1:
        fesc=0;
        switch (n) {
          case 'A': if (cury>1) cury--; break; /* csrup */
          case 'B': if (cury<(nlines-8)) cury++; break; /* csrdn (fixed v1.31) */
          case 'C':
	    if (curx<ncols) curx++;
            else {
              curx=1;
              if (cury<(nlines-8)) cury++;
            }
            break; /* csrr */
          case 'D':
            if (curx>1) curx--;
            else {
	      curx=ncols;
              if (cury>1) cury--;
             }
             break; /* csrl */
          case 'E': clrdsp(); break; /* clrdsp */
          case 'H': curx=1; cury=1; f=1; break; /* home */
          case 'J': /* clrend */
	    for (i=curx;i<ncols;i++) cputs(" "); /* clr only current line */
            if (f) clrdsp(); // for the HP75 27/02/17
            break;
          case 'K': /* clrline */
	    for (i=curx;i<ncols;i++) cputs(" ");
            break;
          case 'Q': fctyp=2; break; /* csrins */
          case 'R': fctyp=1; break; /* csrrep */
          case 'S': /* roll up for demorom75 (v1.31) */
            gotoxy(1,nlines-8);
            cprintf("\n");
            break;
          case '<': fcur=0; break; /* csroff */
          case '>': fcur=1; break; /* csron */
          case '%': fesc=2; break; /* csrpos */
          default: cprintf("Esc%c",n);
        }
        gotoxy(curx,cury);     /* added 1/04/06 for dbg */
        break;
      case 2:
        if (n>=80) n -=64;
        curx=n+1;
        fesc=3;
        break;
      case 3:
        cury=(n&63)+1;
        gotoxy(curx,cury);     /* added 1/04/06 */
        fesc=0;
        break;
    }
  }
  else if (n==8) {
    // needed for the HP75  26/02/17
    if (curx>1) curx--;
    else {
      curx=ncols;
      if (cury>1) cury--;
    }
    gotoxy(curx,cury);
  }
  else {
    switch (n) {
      case   0: n='*'; break;
      case  12: n='u'; break;
      case  28: n='s'; break;
      case  27: fesc=1; break;
      case  29: n='#'; break;
/*      case 124: n='a'; break; */
      case 127: n='`'; break;
      case 10:
      case 13: break;
      default:
//       if ((n<32)||(n>127))  n='.';   // removed for the HP75
       if (n<32)  n='.';
    }
    if (!fesc) {
      // --- added for testing visicalc HP75 (28/02/17) :
      if (n<127)
        textattr(BLACK*16+LIGHTGREEN);
      else
        textattr(GREEN*16+BLACK);
      n &= 127;
      // ---
      cprintf("%c",n);
      curx=wherex();
      cury=wherey();
      f=0;
    }
  }

  update_cursor();

}


