/* ****************************************** */
/* Emu75: HP-75  software emulator for PC     */
/*                                            */
/* pod.c   Expansion POD simulation           */
/*                                            */
/* Version 1.3  J-F Garnier 2018              */
/* ****************************************** */


//#include <mem.h>
#include <stdio.h>
#include <dos.h>


#define GLOBAL extern
#include "capricor.h"

/*
addr   00000...1ffff  20000...fffff  40000...5ffff
memic   ---- 0 ----    ---- 1 ----    ---- 2 ----  ---- 3 ----

addr
0xxxx   built-in HP75 MEM space
1xxxx   N/A
2xxxx   IMEM, 1st 64k space
3xxxx   IMEM, 2nd 64k
4xxxx   XMEM

MEMIC
0    N/A
1    20000-3ffff 64k IMEM, in RAM
     40000-fffff XMEM, in file
2    4xxxx First 64k XMEM, in RAM
     5xxxx to 5dxxx: 2nd 56k XMEM, in file XMEM
     5fxxx   modem interface
3    N/A
*/


static FILE *xmemfp=NULL;


static long bldptr(unsigned char *regptr)
// build a pointer to the e-disc file
{
  long n;

  n= (regptr[2]-2);   // starts from 20000
  n = (n<<8) + regptr[1];
  n = (n<<8) + regptr[0];
  return(n);
}


void pod_close(void)
{
  if (xmemfp!=NULL) {
    fclose(xmemfp);
    xmemfp=NULL;
  }
}


void pod_set(unsigned char *regptr, unsigned char *data, int n)
// set the register A, B, C, D
{
  *regptr++ = data[0];
  if (n>=2) *regptr++ = data[1];
  if (n>=3) *regptr = data[2];
}


void pod_get(unsigned char *regptr, unsigned char *data, int n)
// get the register A, B, C, D, status
{
  data[0]= *regptr++;
  if (n>=2) data[1] = *regptr++;
  if (n>=3) data[2] = *regptr++;
  if (n>=4) data[3] = *regptr;
}

void pod_update(unsigned char *regptr, int n)
// update the register A, B, C, D pointer
{
  int a;

  a=regptr[0]+n;
  regptr[0]=a;
  if (a>255) {
    regptr[1]++;
    if (regptr[1]==0) regptr[2]++;
  }
}



void mod_wrt(void)
// modem write
{
}

void mod_rd(void)
// modem read
{
}


void pod_wrt(unsigned char *regptr, unsigned char *data, int n)
// write to POD memory using register
{
  unsigned char far *ptr;
  int p;
  int ofs;
  long pt;

  // setup pointer
  p=regptr[2];

  if (p<2) return;

  if ((podmemic==2)||(p<=3)) {
    // MEMIC2 or imem, use RAM

    if (p>5) return;
    // map 2xxxx-5xxxx to 32k blocs 0 to 7
    p = (p-2)<<1;  if (regptr[1]&0x80) p++;  // 0 to 7
    ofs=regptr[1]&0x7f;
    ofs = (ofs <<8) + regptr[0];

    if ((podmemic==2)&&(p==7)&&(ofs>=0x6000)) {
      // for modem
      mod_wrt();
    }
    else if ((p==3)&&(ofs>=0x6000)) {
      // gap between imem and xmem needed for initialize ':imem'
    }
    else {
      if (xmem_seg[p]!=0) {
        // write
        ptr= (unsigned char far *) MK_FP(xmem_seg[p],ofs);
        while (n--) *ptr++ = *data++;
      }
    }

  }

  else {
    // MEMIC1 and xmem, use file
    p -=2;
    if ((p*64)>=xmemsize) return;
    pt= bldptr(regptr);
    if (xmemfp==NULL) {
      xmemfp=fopen(xmemfile,"r+b");
    }
    if (xmemfp!=NULL) {
      fseek(xmemfp,pt,0);
      fwrite(data,1,n,xmemfp);
    }
  }


}

void pod_rd(unsigned char *regptr, unsigned char *data, int n)
// read from POD memory using register
{
  unsigned char far *ptr;
  int p;
  int ofs;
  long pt;

  // setup pointer
  p=regptr[2];

  if (p<2) return;

  if ((podmemic==2)||(p<=3)) {
    // MEMIC2 or imem, use RAM

    if (p>5) return;
    // map 2xxxx-5xxxx to 32k blocs 0 to 7
    p = (p-2)<<1; if (regptr[1]&0x80) p++;  // 0 to 7
    ofs=regptr[1]&0x7f;
    ofs = (ofs <<8) + regptr[0];

    if ((podmemic==2)&&(p==7)&&(ofs>=0x6000)) {
    // for modem
  //    mod_rd();
      while (n--) *data++ = 0;    // no modem
    }
    else if ((p==3)&&(ofs>=0x6000)) {
      // gap between imem and xmem needed for initialize ':imem'
      while (n--) *data++ = 0;
    }
    else {
      if (xmem_seg[p]!=0) {
        //  read
        ptr= (unsigned char far *) MK_FP(xmem_seg[p],ofs);
        while (n--) *data++ = *ptr++;
      }
      else {
        while (n--) *data++ = 0;
      }
    }
  }

  else {
    // MEMIC1 and xmem, use file
    p -=2;
    if ((p*64)>=xmemsize) {
      while (n--) *data++ = 0;
      return;
    }
    pt= bldptr(regptr);
    if (xmemfp==NULL) {
      xmemfp=fopen(xmemfile,"r+b");
    }
    if (xmemfp!=NULL) {
      fseek(xmemfp,pt,0);
      fread(data,1,n,xmemfp);
    }
    else {
      while (n--) *data++ = 0;
    }
  }


}


void pod_move(void)
// move memory from A to C for D bytes
// direction given by podsts[0] bit 2 (1=up, 0=down)
{
  long n;
  int pd, ps;
  long pts, ptd;
  unsigned int ofsd, ofss;
  unsigned char far *dest;
  unsigned char far *src;

  ps=podregA[2];
  ofss=podregA[1];
  ofss = (ofss<<8) + podregA[0];

  pd=podregC[2];
  ofsd=podregC[1];
  ofsd = (ofsd<<8) + podregC[0];

  if ((pd==ps)&&(ofsd==ofss)) return;   // nop, used by checksum calc

  if ((ps<2)||(pd<2)) return;

  n=podregD[2];
  n = (n<<8) + podregD[1];
  n = (n<<8) + podregD[0];
  if (n==0) return;

  if ((podmemic==2)||(pd<=3)) {
    // MEMIC2 or imem, use RAM

    if ((ps>5)||(pd>5)) return;

    // map 2xxxx-5xxxx to 32k blocs 0 to 7
    ps = (ps-2)<<1; if (ofss&0x8000) ps++;  // 0 to 7
    ofss &= 0x7fff;
    pd = (pd-2)<<1; if (ofsd&0x8000) pd++;  // 0 to 7
    ofsd &= 0x7fff;

    if (xmem_seg[ps]==0) return;
    if (xmem_seg[pd]==0) return;

    src = MK_FP(xmem_seg[ps],ofss);
    dest= MK_FP(xmem_seg[pd],ofsd);

    if ((podsts[0]&4)==0) {
      // down
      while (n--) {
        *dest-- = *src--;
        if (ofss!=0) {
          ofss--;
        }
        else {
          // change page
          ps--; ofss=0x7fff;
          if (ps<0) return;
          if (xmem_seg[ps]==0) return;
          src = MK_FP(xmem_seg[ps],ofss);
        }
        if (ofsd!=0) {
          ofsd--;
        }
        else {
          // change page
          pd--; ofsd=0x7fff;
          if (pd<0) return;
          if (xmem_seg[pd]==0) return;
          dest = MK_FP(xmem_seg[pd],ofsd);
        }
      } // endwhile
    }
    else {
      //up
      while (n--) {
        *dest++ = *src++;
        ofss++;
        if (ofss&0x8000) {
          // change page
          ps++; ofss=0;
          if (ps>7) return;
          if (xmem_seg[ps]==0) return;
          src = MK_FP(xmem_seg[ps],ofss);
        }
        ofsd++;
        if (ofsd&0x8000) {
          // change page
          pd++; ofsd=0;
          if (pd>7) return;
          if (xmem_seg[pd]==0) return;
          dest = MK_FP(xmem_seg[pd],ofsd);
        }
      } // endwhile
    }
  }

  else {
    // MEMIC1 and xmem, use file
    ps -=2; pd -=2;
    if ( ((ps*64)>=xmemsize) || ((pd*64)>=xmemsize) ) return;

    pts=ps; pts =(pts<<16)+ofss;
    ptd=pd; ptd =(ptd<<16)+ofsd;

    if (xmemfp==NULL) {
      xmemfp=fopen(xmemfile,"r+b");
    }
    if (xmemfp==NULL) return;

    // use global buffer
    if ((podsts[0]&4)==0) {
      // down
      while (n>BUFSIZE) {
        pts-=BUFSIZE;
        ptd-=BUFSIZE;
        fseek(xmemfp,pts,0);
        fread(buffer,1,BUFSIZE,xmemfp);
        fseek(xmemfp,ptd,0);
        fwrite(buffer,1,BUFSIZE,xmemfp);
        n -=BUFSIZE;
      }
      pts-=n;
      ptd-=n;
      fseek(xmemfp,pts,0);
      fread(buffer,1,n,xmemfp);
      fseek(xmemfp,ptd,0);
      fwrite(buffer,1,n,xmemfp);

    }
    else {
      //up
      while (n>BUFSIZE) {
        fseek(xmemfp,pts,0);
        fread(buffer,1,BUFSIZE,xmemfp);
        fseek(xmemfp,ptd,0);
        fwrite(buffer,1,BUFSIZE,xmemfp);
        pts+=BUFSIZE;
        ptd+=BUFSIZE;
        n -=BUFSIZE;
      }
      fseek(xmemfp,pts,0);
      fread(buffer,1,n,xmemfp);
      fseek(xmemfp,ptd,0);
      fwrite(buffer,1,n,xmemfp);
    }

  } // end file


}



void pod_hello(unsigned char *data)
// returns 8 bytes, 2 per MEMIC
{
  data[0]=0; data[1]=0;  // MEMIC0
  data[2]=0; data[3]=0;
  data[4]=0; data[5]=0;
  data[6]=0; data[7]=0;  // MEMIC3
  // MEMIC depending on option:
  if (instmemic==1) {
    data[2]= 0xe3; data[3] = 0x1c;
  }
  if (instmemic) {
    data[4]= 0xe3; data[5] = 0x1c; // MEMIC2  - always there if pod
  }
}

